package RevenueManagement.MultiLeg.BidPriceLP;

import RevenueManagement.MultiLeg.General.*;
import qs.*;

/* number of variables for LP is equal to number of itineraries */
/* number of constraints for LP is equal to number of flights */
/* we regard the demand constraints as simple bounds on the variables */

public class BidPriceLP
{
    /* information related to variables and their constraint coefficients */
    int[][] rmatcnt_var;
    int[][] rmatbeg_var;
    int[][] rmatind_var;
    double[][] rmatval_var;
    double[][] obj_var;
    double[][] lower_var;
    double[][] upper_var;
    String[][] name_var;
    /* information related to the right sides of constraints*/
    double[] rhs_const;
    char[] sense_const;
    int[] varcount_const;
    
    Problem lp;
    static BidPriceLP instance;
    
    public BidPriceLP()
    {
        /* note that the constraint matrix does not change at different time periods */
        /* we generate the constraint matrix before and reuse it */
        /* we do what we did in the qs demo over and over again for each variable in the problem */
        /* note that there is one variable for each itinerary */
        /* the first index of the indices are always reserved for the variable */
        rmatcnt_var = new int[ ItinKeeper.instance().getNoItins() ][ 1 ];
        rmatbeg_var = new int[ ItinKeeper.instance().getNoItins() ][ 1 ];
        /* we do not know how many constraints each itinerary variable appears */
        /* we leave that unspecified for the time being */
        rmatind_var = new int[ ItinKeeper.instance().getNoItins() ][];
        rmatval_var = new double[ ItinKeeper.instance().getNoItins() ][];
        /* the others are as before */
        obj_var = new double[ ItinKeeper.instance().getNoItins() ][ 1 ];
        lower_var = new double[ ItinKeeper.instance().getNoItins() ][ 1 ];
        upper_var = new double[ ItinKeeper.instance().getNoItins() ][ 1 ];
        name_var = new String[ ItinKeeper.instance().getNoItins() ][ 1 ];
        
        rhs_const = new double[ FlightKeeper.instance().getNoFlights() ];
        sense_const = new char[ FlightKeeper.instance().getNoFlights() ];
        /* the length of this array only has to be equal to the number of variables, that's all */
        varcount_const = new int[ ItinKeeper.instance().getNoItins() ];
        
        instance = this;
        generateConstraintMatrix();
        generateConstraints();
    }
    
    public void generateConstraintMatrix()
    {
        /* deal with each variable one by one */
        /* we assume that there is a one-to-one correspondence between the itinerary indices and the variable indices in LP */
        for ( int i = 0 ; i < ItinKeeper.instance().getNoItins() ; i ++ )
        {
            Itin itin = ItinKeeper.instance().getItin( i );
            Flight[] flights = itin.getFlights();
            /* the number of constraints in which this variable appears depend on how many flights are in this itinerary */
            rmatcnt_var[ i ][ 0 ] = flights.length;
            rmatbeg_var[ i ][ 0 ] = 0;
            rmatind_var[ i ] = new int[ flights.length ];
            rmatval_var[ i ] = new double[ flights.length ];
            for ( int f = 0 ; f < flights.length ; f ++ )
            {
                Flight flight = flights[ f ];
                int flindx = FlightKeeper.instance().getFlightIndex( flight );
                rmatind_var[ i ][ f ] = flindx;
                rmatval_var[ i ][ f ] = 1.0;
            }
            obj_var[ i ][ 0 ] = itin.getFare();
            lower_var[ i ][ 0 ] = 0;
            /* set to zero momentarily, this will change later */
            upper_var[ i ][ 0 ] = 0;
            name_var[ i ][ 0 ] = "I_" + itin.getOrigin() + "_" + itin.getDestination() + "_" + itin.getFareClass();
        }
    }
    
    public void generateConstraints()
    {
        /* deal with each constraint one-by-one */
        /* we assume that there is a one-to-one correspondence between the flight indices and the constraint indices in the LP */
        for ( int f = 0 ; f < FlightKeeper.instance().getNoFlights() ; f ++ )
        {
            Flight flight = FlightKeeper.instance().getFlight( f );
            /* set to zero momentarily, this will change later */
            rhs_const[ f ] = 0;
            sense_const[ f ] = 'L';
        }
    }
    
    public void construct ( int time )
    {
        lp = new Problem( "Bid Price LP" );
        lp.change_objsense( QS.MAX );
        
        /* construct the right hand sides of constraints by looking at the capacity keeper */
        for ( int f = 0 ; f < FlightKeeper.instance().getNoFlights() ; f ++ )
        {
            Flight flight = FlightKeeper.instance().getFlight( f );
            rhs_const[ f ] = CapacityKeeper.instance().getCapacity( flight );
        }
        
        /* construct the upper bounds of variables by looking at probability keeper */
        for ( int i = 0 ; i < ItinKeeper.instance().getNoItins() ; i ++ )
        {
            Itin itin = ItinKeeper.instance().getItin( i );
            double p = 0;
            for ( int t = time ; t < GlobalParamsKeeper.instance().getNoTimePeriods() ; t ++ )
            {
                p += ProbabilityKeeper.instance().getProbability( t , itin );
            }
            upper_var[ i ][ 0 ] = p;
        }
        
        try
        {
            /* add constraints to LP */
            lp.add_rows( FlightKeeper.instance().getNoFlights() , varcount_const , null , null , null , rhs_const , sense_const , null);
            
            /* add variables related to each itinerary one-by-one */
            for ( int i = 0 ; i < ItinKeeper.instance().getNoItins() ; i ++ )
            {
                lp.add_cols( 1 , rmatcnt_var[ i ] , rmatbeg_var[ i ] , rmatind_var[ i ] , rmatval_var[ i ] , obj_var[ i ] , lower_var[ i ] , upper_var[ i ] , name_var[ i ] );
            }
        }
        catch ( Exception e )
        {
            throw new Error ( e.getClass() + " " + e.getMessage() );
        }
    }
    
    public void solve()
    {
        try
        { 
            lp.opt_dual();
        }
        catch ( Exception e )
        {
            throw new Error ( e.getClass() + " " + e.getMessage() );
        }
    }
    
    public Problem getLP()
    {
        return lp;
    }
    
    public static BidPriceLP instance()
    {
        return instance;
    }
    
}
