package RevenueManagement.SingleLeg.DP;

import RevenueManagement.SingleLeg.General.*;
import java.io.*;

public class DPSolver
{
    // keeps the value functions
    // the first index keeps the fare class, the second index keeps the capacity
    double[][] vf;
    static DPSolver instance;
    
    public DPSolver()
    {
        instance = this;
        vf = new double[ FareClassKeeper.instance().getNoFareClasses() ][ GlobalParamsKeeper.instance().getCapacity() + 1 ];
    }
    
    
    // solves the dynamic program and stores the value function approximations
    public void solve()
    {
        // loop overall fare classes to compute the value function Vj(.)
        for ( int j = 0 ; j < FareClassKeeper.instance().getNoFareClasses() ; j ++ )
        {
            FareClass fc = FareClassKeeper.instance().getFareClass( j );
            double fare = fc.getFare();
            double[] pdf = fc.getPDF();
            
            // loop over all possible values of remaining capacity to compute the value function Vj(x)
            for ( int x = 0 ; x <= GlobalParamsKeeper.instance().getCapacity() ; x ++ )
            {
                //loop over all possibel values of demand to compute the value function Vj(x,Dj)
                for ( int d = 0 ; d <= fc.getMaxDemand() ; d ++ )
                {
                    double bestValue = computeSaleQuantity( j , x , d );
                    // once found the best value, we know that Vj(x,Dj) = best value
                    // accumulate  Vj(x,Dj) to compute the value function Vk(x)
                    vf[ j ][ x ] = vf[ j ][ x ] + pdf[ d ] * bestValue;
                }
            }
        }
    }
    
    
    // finds how much to sell given the fare class number, remaining  capacity and demand, say z
    // and returns p z + Vj-1(x-z)
    public double computeSaleQuantity ( int fareClassNo , int capacity , int demand )
    {
        FareClass fc = FareClassKeeper.instance().getFareClass( fareClassNo );
        double fare = fc.getFare();
        
        int bestAction = Integer.MAX_VALUE;
        double bestValue = - Double.MAX_VALUE;

        for ( int sales = 0 ; sales <= Math.min( capacity , demand ) ; sales ++ )
        {
            double currentValue = fare * sales;
            // read the value function from next fare class
            double valueFunctionNext;
            // if we are not dealing with the last fare class, 
            // then we can look up the value function for the next fare class
            if ( fareClassNo != 0 )
            {
                valueFunctionNext = vf[ fareClassNo - 1 ][ capacity - sales ];
            }
            else
            {
                valueFunctionNext = 0;
            }
            currentValue = currentValue + valueFunctionNext;
            // check if current value reached is better than the best value
            // if so, then update the best action and best value
            if ( currentValue >= bestValue )
            {
                bestAction = sales;
                bestValue = currentValue;
            }
        }
        return bestValue;
    }
    
    
    // prints the value functions into the file
    // each row corresponds to one fare class j
    public void printValueFunction()
    {
        PrintWriter pw = GlobalParamsKeeper.instance().getOutWriter();
        for ( int j = 0 ; j < FareClassKeeper.instance().getNoFareClasses() ; j ++ )
        {
            for ( int x = 0 ; x <= GlobalParamsKeeper.instance().getCapacity() ; x ++ )
            {
                pw.print( vf[ j ][ x ] + " " );
            }
            pw.println();
        }
        pw.println();
        // flush the print writer so that we can see the output after program terminates
        // and without having to worry about closing the file
        pw.flush();
    }
    
    
    // prints the function pj zj + Vj-1(x-zj) as a function of zj for different values of x
    public void printDecisionFunction()
    {
        PrintWriter pw = GlobalParamsKeeper.instance().getOutWriter();
        for ( int j = 0 ; j < FareClassKeeper.instance().getNoFareClasses() ; j ++ )
        {
            FareClass fc = FareClassKeeper.instance().getFareClass( j );
            for ( int x = 0 ; x <= GlobalParamsKeeper.instance().getCapacity() ; x ++ )
            {
                for ( int z = 0 ; z <= x ; z ++ )
                {
                    double functionValue = fc.getFare() * z;
                    double nextValueFunction;
                    if ( j != 0 )
                    {
                        nextValueFunction = vf[ j - 1 ][ x - z ];
                    }
                    else
                    {
                        nextValueFunction = 0;
                    }
                    functionValue = functionValue + nextValueFunction;
                    pw.print( functionValue + " " );
                }
                pw.println();
            }
            pw.println();
        }
        pw.println();
        pw.flush();
    }
    
    
    public double[][] getValueFunction()
    {
        return vf;
    }
    
    public static DPSolver instance()
    {
        return instance;
    }
            
}
