package RevenueManagement.SingleLeg.DP;

import RevenueManagement.SingleLeg.General.*;
import java.io.*;

public class DPSolverProtection
{
    // keeps the value functions
    // the first index keeps the fare class, the second index keeps the capacity
    double[][] vf;
    // keeps the protection level for each stage and at each capacity level
    // we know that the protection level will not depend on capacity in the end
    // but we keep the protection level corresponding to each capacity any way
    int[][] protections;
    static DPSolverProtection instance;
    
    public DPSolverProtection()
    {
        instance = this;
        vf = new double[ FareClassKeeper.instance().getNoFareClasses() ][ GlobalParamsKeeper.instance().getCapacity() + 1 ];
        protections = new int[ FareClassKeeper.instance().getNoFareClasses() ][ GlobalParamsKeeper.instance().getCapacity() + 1 ];
    }
    
    
    // solves the dynamic program and stores the value function approximations
    public void solve()
    {
        // loop over all fare classes to compute the value function Vj(.)
        for ( int j = 0 ; j < FareClassKeeper.instance().getNoFareClasses() ; j ++ )
        {
            FareClass fc = FareClassKeeper.instance().getFareClass( j );
            double fare = fc.getFare();
            
            // loop over all possible values of remaining capacity to compute the value function Vj(x)
            for ( int x = 0 ; x <= GlobalParamsKeeper.instance().getCapacity() ; x ++ )
            {
                double bestValue = - Double.MAX_VALUE;
                int bestProtectionLevel = -1;
                for ( int y = 0 ; y <= x ; y ++ )
                {
                    double currentValue = fare * computeFirstExpectation( j , x , y ) 
                                                    + computeSecondExpectation( j , x , y );
                    if ( currentValue >= bestValue )
                    {
                        bestValue = currentValue;
                        bestProtectionLevel = y;
                    }
                }
                vf[ j ][ x ] = bestValue;
                protections[ j ][ x ] = bestProtectionLevel;
            }
        }
    }
    
    
    // computes E{min{x-y,D}}
    public double computeFirstExpectation ( int fareClassNo , int capacity , int protection )
    {
        FareClass fc = FareClassKeeper.instance().getFareClass( fareClassNo );
        double[] pdf = fc.getPDF();
        double expectation = 0;
        for ( int d = 0 ; d < pdf.length ; d ++ )
        {
            expectation += pdf[ d ] * Math.min( capacity - protection , d );
        }
        return expectation;
    }
    
    
    // computes E{Vj-1( x - min{x-y,D})}
    public double computeSecondExpectation ( int fareClassNo , int capacity , int protection )
    {
        if ( fareClassNo == 0 )
            return 0;
        
        FareClass fc = FareClassKeeper.instance().getFareClass( fareClassNo );
        double[] pdf = fc.getPDF();
        double expectation = 0;
        for ( int d = 0 ; d < pdf.length ; d ++ )
        {
            expectation += pdf[ d ] * vf[ fareClassNo - 1 ][ capacity - Math.min( capacity - protection , d )];
        }
        return expectation;
    }
    
    
    public int getProtectionLevel ( int fareClassNo , int capacity )
    {
        return protections[ fareClassNo ][ capacity ];
    }
    
    
    // prints the value functions into the file
    // each row corresponds to one fare class j
    public void printValueFunction()
    {
        PrintWriter pw = GlobalParamsKeeper.instance().getOutWriter();
        for ( int j = 0 ; j < FareClassKeeper.instance().getNoFareClasses() ; j ++ )
        {
            for ( int x = 0 ; x <= GlobalParamsKeeper.instance().getCapacity() ; x ++ )
            {
                pw.print( vf[ j ][ x ] + " " );
            }
            pw.println();
        }
        pw.println();
        // flush the print writer so that we can see the output after program terminates
        // and without having to worry about closing the file
        pw.flush();
    }
    
    
    public void printProtectionLevels()
    {
        PrintWriter pw = GlobalParamsKeeper.instance().getOutWriter();
        for ( int j = 0 ; j < FareClassKeeper.instance().getNoFareClasses() ; j ++ )
        {
            for ( int x = 0 ; x <= GlobalParamsKeeper.instance().getCapacity() ; x ++ )
            {
                pw.print( protections[ j ][ x ] + " " );
            }
            pw.println();
        }
        pw.println();
        pw.flush();
    }
    
    
    public double[][] getValueFunction()
    {
        return vf;
    }
    
    public static DPSolverProtection instance()
    {
        return instance;
    }
            
}
